const _ = require('lodash');
const fs = require('fs');
const reactDocgen = require('react-docgen');

function isExampleFile(path) {
  const filename = path.hub.file.opts.filename;
  return /\.example\.jsx?$/.test(filename);
}

function applyExportTransform(path, types, cache) {
  if (!isExampleFile(path)) {
    return;
  }

  const {imports} = cache;
  const properties = path.node.declaration.properties;
  const componentNameProperty = properties.find(function(property) {
    return property.key.name === 'component';
  });
  const componentName = componentNameProperty.value.value;
  const importsKeys = Array.from(imports.keys());
  const componentImport = importsKeys.find(function(filePath) {
    return imports.get(filePath).find(function(specifier) {
      return specifier.local === componentName;
    });
  });
  path.traverse({
    ObjectProperty(path) {
      if (path.node.key.name === 'component') {
        path.insertAfter(
          types.objectProperty(
            types.identifier('path'),
            types.stringLiteral(componentImport)
          )
        );

        const {filePath} = cache;
        if (filePath) {
          path.insertAfter(
            types.objectProperty(
              types.identifier('filePath'),
              types.stringLiteral(filePath)
            )
          );
        }

        const {props} = cache;
        if (props) {
          const flowProps = props.props;
          path.insertAfter(
            types.objectProperty(
              types.identifier('props'),
              types.valueToNode(flowProps)
            )
          );
        }

        path.stop();
      }
    },
  });
}

function applyRenderTransform(path, types, cache) {
  if (!isExampleFile(path)) {
    return;
  }

  const fnNode = path.node;
  if (!fnNode.id || fnNode.id.name !== 'render') {
    return;
  }

  const JSXElements = new Set();
  path.traverse({
    JSXOpeningElement(path) {
      JSXElements.add(path.node.name.name);
    }
  });

  const bodyNode = fnNode.body;
  const source = getSource(path, bodyNode.start, bodyNode.end);
  const blockContent = source.trim().replace(/^{/, '').replace(/}$/, '').trim()
  const prettyBlockContent = prettify(blockContent);

  let implementation = prettyBlockContent;
  if (JSXElements.size > 0) {
    implementation = generateImports(JSXElements, cache) + '\n\n' + prettyBlockContent;
  }

  path.parentPath.insertAfter(
    types.objectProperty(
      types.identifier('implementation'),
      types.stringLiteral(implementation)
    )
  );
}

function prettify(body) {
  const lines = body.split('\n');
  const indentationPerLine = _.compact(lines.map(function(line, index) {
    const indentation = line.match(/^ */);
    return indentation && indentation[0].length;
  }));
  const minimumIndentation = Math.min(...indentationPerLine);

  return lines.map(function(line) {
    return line.replace(new RegExp(`^ {${minimumIndentation}}`), '');
  }).join('\n');
}

function generateImportSpecifier(specifier) {
  if (specifier.imported && specifier.imported !== specifier.local) {
    return `${specifier.imported} as ${specifier.local}`;
  }
  return specifier.local;
}

function generateImports(expressions, cache) {
  const {imports} = cache;
  const generatedImports = [];
  imports.forEach(function(specifiers, filePath) {
    const needImports = specifiers.filter(function(specifier) {
      return expressions.has(specifier.local);
    });
    if (needImports.length > 0) {
      const defaultImport = needImports.find(function(specifier) {
        return specifier.default;
      });
      const otherImports = needImports.filter(function(specifier) {
        return !specifier.default;
      });

      const needImportsArray = [];
      if (defaultImport) {
        needImportsArray.push(generateImportSpecifier(defaultImport));
      }
      const otherImportsArray = [];
      otherImports.forEach(function(specifier) {
        otherImportsArray.push(generateImportSpecifier(specifier));
      });
      if (otherImportsArray.length > 0) {
        needImportsArray.push(`{\n  ${otherImportsArray.join(',\n  ')},\n}`);
      }

      generatedImports.push(`import ${needImportsArray.join(',\n')} from '${filePath}';`);
    }
  });
  return generatedImports.join('\n');
}

function getSource(path, start, end) {
  return path.hub.file.code.slice(start, end);
}

function parseComponent(path) {
  const exampleFile = fs.realpathSync(path.hub.file.opts.filenameRelative);
  const examplePath = exampleFile.split('/');

  const exampleFilename = examplePath[examplePath.length - 1];
  const componentName = exampleFilename.split('.')[0];

  const componentPath = examplePath.slice(0, -1).join('/');

  // Assume that the component module is colocated and named properly.
  const possibleComponentFiles = [
    '/index.js',
    '/index.jsx',
    `/${componentName}.js`,
    `/${componentName}.jsx`,
  ];

  let filePath = null;
  let props = null;

  const componentFile = possibleComponentFiles.find((filename) => {
    filePath = `${componentPath}${filename}`;
    const exists = fs.existsSync(filePath);
    if (!exists) {
      return false;
    }

    try {
      const componentSrc = fs.readFileSync(filePath, 'utf8');
      props = reactDocgen.parse(componentSrc);
      return true;
    } catch (err) {
      // If the file does not define and export its own React component (e.g.
      // an index module that just exports other modules), React Docgen will
      // throw. We can just move on.
      return false;
    }

    return !!props;
  });


  // No colocated component file found for this component.
  if (!props) {
    return {filePath: null, props: null};
  }

  return {filePath, props};
}

module.exports = function({types}) {
  return {
    pre(state) {
      if (!isExampleFile(state.path)) {
        return;
      }

      const imports = new Map();
      state.path.traverse({
        ImportDeclaration(path) {
          path.node.specifiers.forEach(function(specifier) {
            const filepath = path.node.source.value;
            const values = imports.get(filepath) || [];
            values.push({
              default: specifier.type === 'ImportDefaultSpecifier',
              local: specifier.local.name,
              imported: specifier.imported && specifier.imported.name,
            });
            imports.set(filepath, values);
          });
        }
      });

      const {filePath, props} = parseComponent(state.path);

      this.cache = {
        imports,
        filePath,
        props,
      };
    },
    visitor: {
      ExportDefaultDeclaration(path) {
        applyExportTransform(path, types, this.cache);
      },
      ArrowFunctionExpression(path) {
        applyRenderTransform(path, types, this.cache);
      },
      FunctionExpression(path) {
        applyRenderTransform(path, types, this.cache);
      },
    },
  };
}
